@@ -21,7 +21,7 @@ import contextvars
 import json
 
 import structlog
-import toml
+import tomli
 import appdirs
 
 from .lib import nicelogger
@@ -29,11 +29,12 @@ from . import slogconf
 from .util import (
   Entry, Entries, KeyManager, RawResult, Result, VersData,
   FunctionWorker, GetVersionError,
-  FileLoadError,
+  FileLoadError, EntryWaiter,
 )
 from . import __version__
 from .sortversion import sort_version_keys
 from .ctxvars import tries as ctx_tries
+from .ctxvars import entry_waiter as ctx_entry_waiter
 from . import httpclient
 
 logger = structlog.get_logger(logger_name=__name__)
@@ -155,8 +156,9 @@ def load_file(
   use_keymanager: bool,
 ) -> Tuple[Entries, Options]:
   try:
-    config = toml.load(file)
-  except (OSError, toml.TomlDecodeError) as e:
+    with open(file, 'rb') as f:
+      config = tomli.load(f)
+  except (OSError, tomli.TOMLDecodeError) as e:
     raise FileLoadError('version configuration file', file, e)
 
   ver_files: Optional[Tuple[Path, Path]] = None
@@ -219,11 +221,13 @@ class Dispatcher:
     task_sem: asyncio.Semaphore,
     result_q: Queue[RawResult],
     keymanager: KeyManager,
+    entry_waiter: EntryWaiter,
     tries: int,
     source_configs: Dict[str, Dict[str, Any]],
   ) -> List[asyncio.Future]:
     mods: Dict[str, Tuple[types.ModuleType, List]] = {}
     ctx_tries.set(tries)
+    ctx_entry_waiter.set(entry_waiter)
     root_ctx = contextvars.copy_context()
 
     for name, entry in entries.items():
@@ -234,7 +238,7 @@ class Dispatcher:
         mods[source] = mod, tasks
         config = source_configs.get(source)
         if config and getattr(mod, 'configure'):
-          mod.configure(config) # type: ignore
+          mod.configure(config)
       else:
         tasks = mods[source][1]
       tasks.append((name, entry))
@@ -242,7 +246,7 @@ class Dispatcher:
     ret = []
     for mod, tasks in mods.values():
       if hasattr(mod, 'Worker'):
-        worker_cls = mod.Worker # type: ignore
+        worker_cls = mod.Worker
       else:
         worker_cls = FunctionWorker
 
@@ -252,7 +256,7 @@ class Dispatcher:
         task_sem, result_q, tasks, keymanager,
       )
       if worker_cls is FunctionWorker:
-        func = mod.get_version # type: ignore
+        func = mod.get_version
         ctx.run(worker.initialize, func)
 
       ret.append(ctx.run(worker.run))
@@ -307,11 +311,11 @@ def apply_list_options(
 
   sort_version_key = sort_version_keys[
     conf.get("sort_version_key", "parse_version")]
-  versions.sort(key=sort_version_key)
+  versions.sort(key=sort_version_key) # type: ignore
 
   return versions[-1]
 
-def _process_result(r: RawResult) -> Optional[Result]:
+def _process_result(r: RawResult) -> Union[Result, Exception]:
   version = r.version
   conf = r.conf
   name = r.name
@@ -320,11 +324,11 @@ def _process_result(r: RawResult) -> Optional[Result]:
     kw = version.kwargs
     kw['name'] = name
     logger.error(version.msg, **kw)
-    return None
+    return version
   elif isinstance(version, Exception):
     logger.error('unexpected error happened',
                   name=r.name, exc_info=r.version)
-    return None
+    return version
   elif isinstance(version, list):
     version_str = apply_list_options(version, conf)
   else:
@@ -336,10 +340,12 @@ def _process_result(r: RawResult) -> Optional[Result]:
     try:
       version_str = substitute_version(version_str, conf)
       return Result(name, version_str, conf)
-    except (ValueError, re.error):
+    except (ValueError, re.error) as e:
       logger.exception('error occurred in version substitutions', name=name)
+      return e
 
-  return None
+  else:
+    return ValueError('no version returned')
 
 def check_version_update(
   oldvers: VersData, name: str, version: str,
@@ -353,18 +359,27 @@ def check_version_update(
 async def process_result(
   oldvers: VersData,
   result_q: Queue[RawResult],
-) -> VersData:
+  entry_waiter: EntryWaiter,
+) -> Tuple[VersData, bool]:
   ret = {}
+  has_failures = False
   try:
     while True:
       r = await result_q.get()
-      r1 = _process_result(r)
-      if r1 is None:
+      try:
+        r1 = _process_result(r)
+      except Exception as e:
+        logger.exception('error processing result', result=r)
+        r1 = e
+      if isinstance(r1, Exception):
+        entry_waiter.set_exception(r.name, r1)
+        has_failures = True
         continue
       check_version_update(oldvers, r1.name, r1.version)
+      entry_waiter.set_result(r1.name, r1.version)
       ret[r1.name] = r1.version
   except asyncio.CancelledError:
-    return ret
+    return ret, has_failures
 
 async def run_tasks(
   futures: Sequence[Awaitable[None]]
